import UIKit

public final class IsDisplayedAndMatchesCheck: ElementInteraction {
    private let overridenPercentageOfVisibleArea: CGFloat?
    private let buildMatcher: (ElementMatcherBuilder) -> ElementMatcher
    
    public init(
        overridenPercentageOfVisibleArea: CGFloat?,
        buildMatcher: @escaping (ElementMatcherBuilder) -> ElementMatcher)
    {
        self.overridenPercentageOfVisibleArea = overridenPercentageOfVisibleArea
        self.buildMatcher = buildMatcher
    }
    
    public convenience init(
        overridenPercentageOfVisibleArea: CGFloat?,
        matcher: ElementMatcher)
    {
        self.init(
            overridenPercentageOfVisibleArea: overridenPercentageOfVisibleArea,
            buildMatcher: { _ in matcher }
        )
    }
    
    public func with(
        dependencies: ElementInteractionDependencies)
        -> ElementInteractionWithDependencies
    {
        return WithDependencies(
            dependencies: dependencies,
            overridenPercentageOfVisibleArea: overridenPercentageOfVisibleArea,
            buildMatcher: buildMatcher
        )
    }
    
    public final class WithDependencies: ElementInteractionWithDependencies {
        private let dependencies: ElementInteractionDependencies
        private let overridenPercentageOfVisibleArea: CGFloat?
        private let buildMatcher: (ElementMatcherBuilder) -> ElementMatcher
        
        public init(
            dependencies: ElementInteractionDependencies,
            overridenPercentageOfVisibleArea: CGFloat?,
            buildMatcher: @escaping (ElementMatcherBuilder) -> ElementMatcher)
        {
            self.dependencies = dependencies
            self.overridenPercentageOfVisibleArea = overridenPercentageOfVisibleArea
            self.buildMatcher = buildMatcher
        }
        
        public func description() -> String {
            """
            element "\(dependencies.elementInfo.elementName)" matches matcher "\(buildMatcher(dependencies.elementMatcherBuilder).description)"
            """
        }
        
        public func interactionFailureShouldStopTest() -> Bool {
            return false
        }
        
        public func perform() -> InteractionResult {
            return dependencies.interactionRetrier.retryInteractionUntilTimeout { [buildMatcher, dependencies] _ in
                dependencies.interactionResultMaker.makeResultCatchingErrors {
                    try dependencies.snapshotResolver.resolve(overridenPercentageOfVisibleArea: overridenPercentageOfVisibleArea) { snapshot in
                        let matcher = buildMatcher(dependencies.elementMatcherBuilder)
                        
                        switch matcher.match(value: snapshot) {
                        case .match:
                            return .success
                        case let .mismatch(mismatchResult):
                            return dependencies.interactionResultMaker.failure(
                                message: "check failed (\(matcher.description)): \(mismatchResult.mismatchDescription)",
                                attachments: mismatchResult.attachments
                            )
                        }
                    }
                }
            }
        }
    }
}
